//
//  ClassRoomCommand.swift
//  Flat
//
//  Created by xuyunshi on 2021/11/15.
//  Copyright © 2021 agora.io. All rights reserved.
//

import Foundation
import AgoraRtmKit
import RxSwift
import RxCocoa

class RtmChannel: NSObject, AgoraRtmChannelDelegate {
    let newMemberPublisher: PublishRelay<String> = .init()
    let memberLeftPublisher: PublishRelay<String> = .init()
    let newMessagePublish: PublishRelay<(text: String, date: Date, sender: String)> = .init()
    let rawDataPublish: PublishRelay<(data: Data, sender: String)> = .init()
    
    var userUUID: String!
    var channelId: String!
    weak var channel: AgoraRtmChannel!
    
    deinit {
        logger.trace("\(self), channelId \(channelId ?? "") deinit")
    }
    
    func sendRawData(_ data: Data) -> Single<Void> {
        .create { [weak self] observer in
            guard let self = self else {
                observer(.failure("self not exist"))
                return Disposables.create()
            }
            let msg = AgoraRtmRawMessage(rawData: data, description: "")
            self.channel.send(msg) { error in
                if error == .errorOk {
                    observer(.success(()))
                } else {
                    observer(.failure("send message error \(error)"))
                }
            }
            return Disposables.create()
        }
    }
    
    func sendMessage(_ text: String, censor: Bool = false, appendToNewMessage: Bool = false) -> Single<Void> {
        let send = Single<Void>.create { [weak self] observer in
            guard let self = self else {
                observer(.failure("self not exist"))
                return Disposables.create()
            }
            
            self.channel.send(.init(text: text)) { error in
                if error == .errorOk {
                    observer(.success(()))
                } else {
                    observer(.failure("send message error \(error)"))
                }
            }
            return Disposables.create()
        }.do(onSuccess: { [weak self] in
            guard let self = self else { return }
            if appendToNewMessage {
                self.newMessagePublish.accept((text, Date(), self.userUUID))
            }
        })
        if censor {
            return ApiProvider.shared.request(fromApi: MessageCensorRequest(text: text))
                .asSingle()
                .flatMap { r in return r.valid ? send : .just(()) }
        } else {
            return send
        }
    }
    
    func getMembers() -> Single<[String]> {
        return .create { [weak self] observer in
            guard let self = self else {
                observer(.failure("self not exist"))
                return Disposables.create()
            }
            logger.info("start get members")
            self.channel.getMembersWithCompletion { members, error in
                guard error == .ok else {
                    let strError = "get member error, \(error.rawValue)"
                    observer(.failure(strError))
                    logger.error("\(strError)")
                    return
                }
                let memberIds = members?.map { $0.userId } ?? []
                logger.info("success get members \(memberIds)")
                observer(.success(memberIds))
            }
            return Disposables.create()
        }
    }
    
    func channel(_ channel: AgoraRtmChannel, memberJoined member: AgoraRtmMember) {
        logger.info("memberJoined \(member.userId)")
        newMemberPublisher.accept(member.userId)
    }
    
    func channel(_ channel: AgoraRtmChannel, memberLeft member: AgoraRtmMember) {
        logger.info("memberLeft \(member.userId)")
        memberLeftPublisher.accept(member.userId)
    }
    
    func channel(_ channel: AgoraRtmChannel, messageReceived message: AgoraRtmMessage, from member: AgoraRtmMember) {
//        logger.info("receive \(type)
        switch message.type {
        case .text:
            newMessagePublish.accept((message.text, Date(timeIntervalSince1970: TimeInterval(message.serverReceivedTs)), member.userId))
        case .raw:
            if let rawMessage = message as? AgoraRtmRawMessage {
                rawDataPublish.accept((rawMessage.rawData, member.userId))
            }
        default:
            return
        }
    }
}

extension AgoraRtmMessageType: CustomStringConvertible {
    public var description: String {
        switch self {
        case .undefined:
            return "undefined"
        case .text:
            return "text"
        case .raw:
            return "raw"
        case .file:
            return "file"
        case .image:
            return "image"
        @unknown default:
            return "undefined"
        }
    }
}
